@@ -1,17 +1,20 @@
+import datetime
+import importlib
 import logging
 import os
 import sys
 import time
+from collections import Counter
 
 import pytest
 from sentry_sdk.client import Client
-
 from tests.conftest import patch_start_tracing_child
 
 import sentry_sdk
+import sentry_sdk.scope
 from sentry_sdk import (
+    get_client,
     push_scope,
-    configure_scope,
     capture_event,
     capture_exception,
     capture_message,
@@ -19,20 +22,19 @@ from sentry_sdk import (
     last_event_id,
     add_breadcrumb,
     isolation_scope,
+    new_scope,
     Hub,
-    Scope,
 )
 from sentry_sdk.integrations import (
     _AUTO_ENABLING_INTEGRATIONS,
+    _DEFAULT_INTEGRATIONS,
     Integration,
     setup_integrations,
 )
 from sentry_sdk.integrations.logging import LoggingIntegration
 from sentry_sdk.integrations.redis import RedisIntegration
-from sentry_sdk.scope import (  # noqa: F401
-    add_global_event_processor,
-    global_event_processors,
-)
+from sentry_sdk.integrations.stdlib import StdlibIntegration
+from sentry_sdk.scope import add_global_event_processor
 from sentry_sdk.utils import get_sdk_name, reraise
 from sentry_sdk.tracing_utils import has_tracing_enabled
 
@@ -71,13 +73,11 @@ def test_processors(sentry_init, capture_events):
     sentry_init()
     events = capture_events()
 
-    with configure_scope() as scope:
+    def error_processor(event, exc_info):
+        event["exception"]["values"][0]["value"] += " whatever"
+        return event
 
-        def error_processor(event, exc_info):
-            event["exception"]["values"][0]["value"] += " whatever"
-            return event
-
-        scope.add_error_processor(error_processor, ValueError)
+    sentry_sdk.get_isolation_scope().add_error_processor(error_processor, ValueError)
 
     try:
         raise ValueError("aha!")
@@ -294,7 +294,7 @@ def test_breadcrumb_arguments(sentry_init, capture_events):
     add_breadcrumb(crumb=dict(foo=42))
 
 
-def test_push_scope(sentry_init, capture_events):
+def test_push_scope(sentry_init, capture_events, suppress_deprecation_warnings):
     sentry_init()
     events = capture_events()
 
@@ -311,7 +311,9 @@ def test_push_scope(sentry_init, capture_events):
     assert "exception" in event
 
 
-def test_push_scope_null_client(sentry_init, capture_events):
+def test_push_scope_null_client(
+    sentry_init, capture_events, suppress_deprecation_warnings
+):
     """
     This test can be removed when we remove push_scope and the Hub from the SDK.
     """
@@ -385,22 +387,53 @@ def test_breadcrumbs(sentry_init, capture_events):
             category="auth", message="Authenticated user %s" % i, level="info"
         )
 
-    Scope.get_isolation_scope().clear()
+    sentry_sdk.get_isolation_scope().clear()
 
     capture_exception(ValueError())
     (event,) = events
     assert len(event["breadcrumbs"]["values"]) == 0
 
 
+def test_breadcrumb_ordering(sentry_init, capture_events):
+    sentry_init()
+    events = capture_events()
+
+    timestamps = [
+        datetime.datetime.now() - datetime.timedelta(days=10),
+        datetime.datetime.now() - datetime.timedelta(days=8),
+        datetime.datetime.now() - datetime.timedelta(days=12),
+    ]
+
+    for timestamp in timestamps:
+        add_breadcrumb(
+            message="Authenticated at %s" % timestamp,
+            category="auth",
+            level="info",
+            timestamp=timestamp,
+        )
+
+    capture_exception(ValueError())
+    (event,) = events
+
+    assert len(event["breadcrumbs"]["values"]) == len(timestamps)
+    timestamps_from_event = [
+        datetime.datetime.strptime(
+            x["timestamp"].replace("Z", ""), "%Y-%m-%dT%H:%M:%S.%f"
+        )
+        for x in event["breadcrumbs"]["values"]
+    ]
+    assert timestamps_from_event == sorted(timestamps)
+
+
 def test_attachments(sentry_init, capture_envelopes):
     sentry_init()
     envelopes = capture_envelopes()
 
     this_file = os.path.abspath(__file__.rstrip("c"))
 
-    with configure_scope() as scope:
-        scope.add_attachment(bytes=b"Hello World!", filename="message.txt")
-        scope.add_attachment(path=this_file)
+    scope = sentry_sdk.get_isolation_scope()
+    scope.add_attachment(bytes=b"Hello World!", filename="message.txt")
+    scope.add_attachment(path=this_file)
 
     capture_exception(ValueError())
 
@@ -425,6 +458,21 @@ def test_attachments(sentry_init, capture_envelopes):
         assert pyfile.payload.get_bytes() == f.read()
 
 
+@pytest.mark.tests_internal_exceptions
+def test_attachments_graceful_failure(
+    sentry_init, capture_envelopes, internal_exceptions
+):
+    sentry_init()
+    envelopes = capture_envelopes()
+
+    sentry_sdk.get_isolation_scope().add_attachment(path="non_existent")
+    capture_exception(ValueError())
+
+    (envelope,) = envelopes
+    assert len(envelope.items) == 2
+    assert envelope.items[1].payload.get_bytes() == b""
+
+
 def test_integration_scoping(sentry_init, capture_events):
     logger = logging.getLogger("test_basics")
 
@@ -442,6 +490,51 @@ def test_integration_scoping(sentry_init, capture_events):
     assert not events
 
 
+default_integrations = [
+    getattr(
+        importlib.import_module(integration.rsplit(".", 1)[0]),
+        integration.rsplit(".", 1)[1],
+    )
+    for integration in _DEFAULT_INTEGRATIONS
+]
+
+
+@pytest.mark.forked
+@pytest.mark.parametrize(
+    "provided_integrations,default_integrations,disabled_integrations,expected_integrations",
+    [
+        ([], False, None, set()),
+        ([], False, [], set()),
+        ([LoggingIntegration()], False, None, {LoggingIntegration}),
+        ([], True, None, set(default_integrations)),
+        (
+            [],
+            True,
+            [LoggingIntegration(), StdlibIntegration],
+            set(default_integrations) - {LoggingIntegration, StdlibIntegration},
+        ),
+    ],
+)
+def test_integrations(
+    sentry_init,
+    provided_integrations,
+    default_integrations,
+    disabled_integrations,
+    expected_integrations,
+    reset_integrations,
+):
+    sentry_init(
+        integrations=provided_integrations,
+        default_integrations=default_integrations,
+        disabled_integrations=disabled_integrations,
+        auto_enabling_integrations=False,
+        debug=True,
+    )
+    assert {
+        type(integration) for integration in get_client().integrations.values()
+    } == expected_integrations
+
+
 @pytest.mark.skip(
     reason="This test is not valid anymore, because with the new Scopes calling bind_client on the Hub sets the client on the global scope. This test should be removed once the Hub is removed"
 )
@@ -515,14 +608,14 @@ def test_scope_event_processor_order(sentry_init, capture_events):
     sentry_init(debug=True, before_send=before_send)
     events = capture_events()
 
-    with push_scope() as scope:
+    with new_scope() as scope:
 
         @scope.add_event_processor
         def foo(event, hint):
             event["message"] += "foo"
             return event
 
-        with push_scope() as scope:
+        with new_scope() as scope:
 
             @scope.add_event_processor
             def bar(event, hint):
@@ -546,7 +639,7 @@ def test_capture_event_with_scope_kwargs(sentry_init, capture_events):
 
 
 def test_dedupe_event_processor_drop_records_client_report(
-    sentry_init, capture_events, capture_client_reports
+    sentry_init, capture_events, capture_record_lost_event_calls
 ):
     """
     DedupeIntegration internally has an event_processor that filters duplicate exceptions.
@@ -555,7 +648,7 @@ def test_dedupe_event_processor_drop_records_client_report(
     """
     sentry_init()
     events = capture_events()
-    reports = capture_client_reports()
+    record_lost_event_calls = capture_record_lost_event_calls()
 
     try:
         raise ValueError("aha!")
@@ -567,35 +660,50 @@ def test_dedupe_event_processor_drop_records_client_report(
             capture_exception()
 
     (event,) = events
-    (report,) = reports
+    (lost_event_call,) = record_lost_event_calls
 
     assert event["level"] == "error"
     assert "exception" in event
-    assert report == ("event_processor", "error")
+    assert lost_event_call == ("event_processor", "error", None, 1)
 
 
 def test_event_processor_drop_records_client_report(
-    sentry_init, capture_events, capture_client_reports
+    sentry_init, capture_events, capture_record_lost_event_calls
 ):
     sentry_init(traces_sample_rate=1.0)
     events = capture_events()
-    reports = capture_client_reports()
+    record_lost_event_calls = capture_record_lost_event_calls()
 
-    global global_event_processors
+    # Ensure full idempotency by restoring the original global event processors list object, not just a copy.
+    old_processors = sentry_sdk.scope.global_event_processors
 
-    @add_global_event_processor
-    def foo(event, hint):
-        return None
+    try:
+        sentry_sdk.scope.global_event_processors = (
+            sentry_sdk.scope.global_event_processors.copy()
+        )
 
-    capture_message("dropped")
+        @add_global_event_processor
+        def foo(event, hint):
+            return None
 
-    with start_transaction(name="dropped"):
-        pass
+        capture_message("dropped")
 
-    assert len(events) == 0
-    assert reports == [("event_processor", "error"), ("event_processor", "transaction")]
+        with start_transaction(name="dropped"):
+            pass
 
-    global_event_processors.pop()
+        assert len(events) == 0
+
+        # Using Counter because order of record_lost_event calls does not matter
+        assert Counter(record_lost_event_calls) == Counter(
+            [
+                ("event_processor", "error", None, 1),
+                ("event_processor", "transaction", None, 1),
+                ("event_processor", "span", None, 1),
+            ]
+        )
+
+    finally:
+        sentry_sdk.scope.global_event_processors = old_processors
 
 
 @pytest.mark.parametrize(
@@ -825,3 +933,21 @@ def test_last_event_id_scope(sentry_init):
     # Should not crash
     with isolation_scope() as scope:
         assert scope.last_event_id() is None
+
+
+def test_hub_constructor_deprecation_warning():
+    with pytest.warns(sentry_sdk.hub.SentryHubDeprecationWarning):
+        Hub()
+
+
+def test_hub_current_deprecation_warning():
+    with pytest.warns(sentry_sdk.hub.SentryHubDeprecationWarning) as warning_records:
+        Hub.current
+
+    # Make sure we only issue one deprecation warning
+    assert len(warning_records) == 1
+
+
+def test_hub_main_deprecation_warnings():
+    with pytest.warns(sentry_sdk.hub.SentryHubDeprecationWarning):
+        Hub.main
